from django.db.models import FileField, ImageField

from .storages import private_media_storage


class PrivateMediaFileField(FileField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("storage", private_media_storage)
        super().__init__(*args, **kwargs)


class PrivateMediaImageField(ImageField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("storage", private_media_storage)
        super().__init__(*args, **kwargs)



from django.contrib.auth.mixins import PermissionRequiredMixin
from django.views.generic import DetailView

from django_sendfile import sendfile


class PrivateMediaView(PermissionRequiredMixin, DetailView):
    """
    Verify the permission required and send the filefield via sendfile.

    :param permission_required: the permission required to view the file
    :param model: the model class to look up the object
    :param file_field: the name of the ``Filefield``
    """

    file_field = None
    # see :func:`sendfile.sendfile` for available parameters
    sendfile_options = None

    def get_sendfile_opts(self):
        return self.sendfile_options or {}

    def get(self, request, *args, **kwargs):
        filename = getattr(self.get_object(), self.file_field).path
        sendfile_options = self.get_sendfile_opts()
        return sendfile(request, filename, **sendfile_options)



from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.utils.functional import LazyObject, cached_property


class PrivateMediaFileSystemStorage(FileSystemStorage):
    """
    Storage that puts files in the private media folder that isn't
    globally available.
    """

    def _clear_cached_properties(self, setting, **kwargs):
        super()._clear_cached_properties(setting, **kwargs)
        if setting == "PRIVATE_MEDIA_ROOT":
            self.__dict__.pop("base_location", None)
            self.__dict__.pop("location", None)
        elif setting == "PRIVATE_MEDIA_URL":
            self.__dict__.pop("base_url", None)

    @cached_property
    def base_location(self):
        return self._value_or_setting(self._location, settings.PRIVATE_MEDIA_ROOT)

    @cached_property
    def base_url(self):
        if self._base_url is not None and not self._base_url.endswith("/"):
            self._base_url += "/"
        return self._value_or_setting(self._base_url, settings.PRIVATE_MEDIA_URL)


class PrivateMediaStorage(LazyObject):
    def _setup(self):
        self._wrapped = PrivateMediaFileSystemStorage()


private_media_storage = PrivateMediaStorage()



from django.contrib.auth import get_permission_codename
from django.urls import re_path

from .views import PrivateMediaView
from .widgets import PrivateFileWidget


class PrivateMediaMixin:
    private_media_fields = ()
    private_media_permission_required = None
    private_media_view_class = PrivateMediaView
    private_media_file_widget = PrivateFileWidget
    # options passed through to sendfile, as a dict
    private_media_view_options = None

    def get_private_media_fields(self):
        return self.private_media_fields

    def get_private_media_permission_required(self, field: str):
        if self.private_media_permission_required:
            return self.private_media_permission_required

        opts = self.opts
        codename = get_permission_codename("change", opts)
        return "%s.%s" % (opts.app_label, codename)

    def get_private_media_view_options(self, field):
        return self.private_media_view_options or {}

    def get_private_media_view(self, field):
        View = self.private_media_view_class
        return View.as_view(
            model=self.model,
            file_field=field,
            permission_required=self.get_private_media_permission_required(field),
            sendfile_options=self.get_private_media_view_options(field),
        )

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        field = super().formfield_for_dbfield(db_field, request, **kwargs)
        private_media_fields = self.get_private_media_fields()
        if db_field.name in private_media_fields:
            # replace the widget
            view_name = self._get_private_media_view_name(db_field.name)
            # TODO: don't nuke potential other overrides?
            Widget = self.private_media_file_widget
            field.widget = Widget(url_name="admin:%s" % view_name)
        return field

    def _get_private_media_view_name(self, field):
        name = "%(app_label)s_%(model_name)s_%(field)s" % {
            "app_label": self.opts.app_label,
            "model_name": self.opts.model_name,
            "field": field,
        }
        return name

    def get_urls(self):
        default = super().get_urls()

        extra = []
        for field in self.get_private_media_fields():
            view = self.get_private_media_view(field)
            extra.append(
                re_path(
                    r"^(?P<pk>\d+)/%s/$" % field,
                    self.admin_site.admin_view(view),
                    name=self._get_private_media_view_name(field),
                )
            )

        return extra + default



import tempfile

from django.test import override_settings


def temp_private_root():
    # Convenience decorator/context manager to use a temporary directory as
    # PRIVATE_MEDIA_ROOT.
    tmpdir = tempfile.mkdtemp()
    return override_settings(PRIVATE_MEDIA_ROOT=tmpdir, SENDFILE_ROOT=tmpdir)



from typing import Optional, Union

from django.contrib.admin.widgets import AdminFileWidget
from django.core.files.uploadedfile import UploadedFile
from django.db.models.fields.files import FieldFile
from django.urls import reverse


class PrivateFileWidgetMixin:
    def __init__(self, *args, **kwargs):
        self.url_name = kwargs.pop("url_name")
        super().__init__(*args, **kwargs)

    def get_context(
        self, name: str, value: Optional[Union[FieldFile, UploadedFile]], attrs: dict
    ):
        """
        Return value-related substitutions.
        """
        context = super().get_context(name, value, attrs)
        if self.is_initial(value):
            context["url"] = reverse(self.url_name, kwargs={"pk": value.instance.pk})
            context["display_value"] = self.get_display_value(value)
        return context

    def get_display_value(self, value):
        return value


class PrivateFileWidget(PrivateFileWidgetMixin, AdminFileWidget):
    template_name = "admin/widgets/clearable_private_file_input.html"
